subseqs :: [a] -> [[a]]
subseqs (x:xs)	= xss ++ [x] : (map (x:) xss)
  where xss = subseqs xs
subseqs _ 	= []

data Expr = Num Int
	  | App Op Expr Expr
	    deriving (Show)
	  
data Op = Add
	| Sub
	| Mul
	| Div
	  deriving (Show, Eq)
	
type Value = Int

value :: Expr -> Value
value (Num x)		= x
value (App op e1 e2)	= apply op (value e1) (value e2)

apply Add = (+)
apply Sub = (-)
apply Mul = (*)
apply Div = div

non :: Op -> Expr -> Bool
non op (Num x)		= True
non op1 (App op2 e1 e2)	= not (op1 == op2)


{- legal Mul (e1, v 1) (e2, v 2)
= (1 < v 1 ∧ v 1 ≤ v 2) ∧ non Div e1 ∧ non Mul e2 ∧ non Div e2
 -}
 

{-
legal Add (e1,v1) (e2,v2)	= v1 <= v2
legal Sub (e1,v1) (e2,v2)	= v1 > v2
legal Mul (e1,v1) (e2,v2)	= (1 < v1) && (v1 <= v2)
legal Div (e1,v1) (e2,v2)	= (1 < v2) && (mod v1 v2 == 0)
-}

legal :: Op -> (Expr, Value) -> (Expr, Value) -> Bool
legal Add (e1,v1) (e2,v2)	= (v1 <= v2) && (non Sub e1) && (non Add e2) && (non Sub e2)
legal Sub (e1,v1) (e2,v2)	= (v1 > v2) && (non Sub e1) && (non Sub e2)
legal Mul (e1,v1) (e2,v2)	= (1 < v1) && (v1 <= v2) && (non Div e1) && (non Mul e1) && (non Div e2)
-- legal Mul (e1,v1) (e2,v2)	= (1 < v1) && (v1 <= v2) && (non Div e1)  && (non Div e2)
legal Div (e1,v1) (e2,v2)	= (1 < v2) && (mod v1 v2 == 0) && (non Div e1) && (non Div e2)
  


mkExprs :: [Int] -> [(Expr, Value)]
mkExprs [x]	= [(Num x, x)]
mkExprs xs	= [ev | (ys,zs) <- unmerges xs,
			ev1 <- mkExprs ys,
			ev2 <- mkExprs zs,
			ev <- combine ev1 ev2]
			
unmerges :: [a] -> [([a], [a])]
unmerges [x,y]	= [([x], [y])]
unmerges (x:xs)	= [([x], xs)] ++
		      concatMap (add x) (unmerges xs)
  where add x (ys, zs) = [(x: ys, zs), (ys, x:zs)]
	
	
	
{-
combine :: (Expr, Value) -> (Expr, Value) -> [(Expr, Value)]
combine (e1, v1) (e2, v2)
  | v1 < v2	= comb1 (e1,v1) (e2,v2)
  | v1 == v2	= comb2 (e1,v1) (e2,v2)
  | v1 > v2	= comb1 (e2,v2) (e1,v1)

comb1 (e1,v1) (e2,v2)
  = (if (non Sub e1) && (non Sub e2) then
      [(App Add e1 e2, v1 + v2) | non Add e2] ++ [(App Sub e2 e1, v2 - v1)]
    else []) ++
    (if (1 < v1) && (non Div e1) && (non Div e2) then
      [(App Mul  e1 e2, v1 * v2) | non Mul e2] ++ [(App Div e2 e1, q) | r == 0]
    else [])
  where (q,r) = divMod v2 v1
	
comb2 (e1,v1) (e2,v2)
  = [(App Add e1 e2, v1 + v2) | non Sub e1, non Add e2, non Sub e2] ++
    (if (1 < v1) && (non Div e1) && (non Div e2) then
      [(App Mul e1 e2, v1 * v2) | non Mul e2] ++ [(App Div e1 e2, 1)]
    else [])
      -}
combine :: (Expr, Value) -> (Expr, Value) -> [(Expr,Value)]
combine (e1,v1)(e2,v2) = 
			  [(App op e1 e2, apply op v1 v2) | op <- ops, legal op (e1,v1) (e2,v2)] ++
                          [(App op e2 e1, apply op v2 v1) | op <- ops, legal op (e2,v2) (e1,v1)]
                         where ops = [Add,Sub,Mul,Div]
			       
nearest n ((e,v): evs)	= if d == 0 then (e,v)
				   else search n d (e,v) evs
  where d = abs (n - v)
	
search n d ev []	= ev
search n d ev ((e,v): evs) 
  | d' == 0	= (e,v)
  | d' < d	= search n d' (e,v) evs
  | d' >= d	= search n d ev evs
  where d' = abs (n - v)
			
			
countdown :: Int -> [Int] -> (Expr, Value)
countdown n = nearest n . concatMap mkExprs . subseqs

display :: (Expr, Value) -> String
display (e, v)	= (display' e) ++ " = " ++ show v

display' :: Expr -> String
display' (Num x)	= show x
display' (App op e1 e2)	= "(" ++ (display' e1) ++ (disOp op) ++ (display' e2) ++ ")"
  where disOp Add = " + "
	disOp Sub = " - "
	disOp Mul = " * "
	disOp Div = " / "
	
exprs :: [Int] -> [(Expr, Value)]	
exprs = concatMap mkExprs . subseqs 

cdp :: Int -> [Int] -> IO()
cdp n ns = print (display (countdown n ns))

--main = print (display (countdown 83100 [1,2,3,4,5,6,7,8]))
--main = print (display (countdown 831 [1,3,7,10,25,50]))