subseqs :: [Int] -> [[Int]]
subseqs (x:xs)	= xss ++ [x] : (map (x:) xss)
  where xss = subseqs xs
subseqs _ 	= []

data Expr = Num Int
	  | App Op Expr Expr
	    deriving (Show)
	  
data Op = Add
	| Sub
	| Mul
	| Div
	  deriving (Show, Eq)
	
type Value = Int

data Trie a = Node a [(Int, Trie a)]
  deriving (Show)

data Tree = Tip Int | Bin Tree Tree 
  deriving (Show)

type Memo = Trie [Tree]

empty :: Memo
empty = Node [] []

mkTrees :: Memo -> [Int] -> [Tree]
mkTrees memo [x]	= [Tip x]
mkTrees memo xs		= [Bin t1 t2 | (ys, zs) <- unmerges xs,
					t1 <- fetch memo ys,
					t2 <- fetch memo zs]

toExprs :: Tree -> [(Expr, Value)]
toExprs (Tip x) 	= [(Num x, x)]
toExprs (Bin t1 t2)	= [ev | ev1 <- toExprs t1, ev2 <- toExprs t2, ev <- combine ev1 ev2]

--fetch :: Memo -> [Int] -> [(Expr, Value)]
fetch (Node es xms) []		= es
fetch (Node es xms) (x:xs)	= fetch (follow x xms) xs

follow :: Int -> [(Int, Memo)] -> Memo
follow x xms = head [m | (x', m) <- xms, x == x']

--store :: [Int] -> [(Expr, Value)] -> Memo -> Memo
store [x] es (Node fs xms)	= Node fs ((x, Node es []) : xms)
store (x:xs) es (Node fs xms)	= Node fs (yms ++ (x, store xs es m) : zms)
  where (yms, (z,m) : zms) = break (equals x) xms
	equals x (z,m) = x == z
	
--extract :: Memo -> [(Expr, Value)]
extract (Node es xms) = es ++ concatMap (extract . snd) xms

value :: Expr -> Value
value (Num x)		= x
value (App op e1 e2)	= apply op (value e1) (value e2)

apply Add = (+)
apply Sub = (-)
apply Mul = (*)
apply Div = div
			
unmerges :: [a] -> [([a], [a])]
--unmerges [] 	= []
unmerges [x,y]	= [([x], [y])]
unmerges (x:xs)	= [([x], xs)] ++
		      concatMap (add x) (unmerges xs)
  where add x (ys, zs) = [(x: ys, zs), (ys, x:zs)]
	
--combine :: (Expr, Value) -> (Expr, Value) -> [(Expr, Value)]
combine (e1, v1) (e2, v2)
  | v1 < v2	= comb1 (e1,v1) (e2,v2)
  | v1 == v2	= comb2 (e1,v1) (e2,v2)
  | v1 > v2	= comb1 (e2,v2) (e1,v1)
  
comb1 (e1,v1) (e2,v2)
  = [(App Add e1 e2, v1 + v2), (App Sub e2 e1, v2 -v1)] ++
    if 1 < v1 then
    [(App Mul  e1 e2, v1 * v2)] ++ [(App Div e2 e1, q) | r == 0]
    else []
  where (q,r) = divMod v2 v1
	
comb2 (e1,v1) (e2,v2)
  = [(App Add e1 e2, v1 + v2)] ++
    if 1 < v1 then
    [(App Mul e1 e2, v1 * v2), (App Div e1 e2, 1)]
    else []
  
nearest n ((e,v): evs)	= if d == 0 then (e,v)
				   else search n d (e,v) evs
  where d = abs (n - v)
	
search n d ev []	= ev
search n d ev ((e,v): evs) 
  | d' == 0	= (e,v)
  | d' < d	= search n d' (e,v) evs
  | d' >= d	= search n d ev evs
  where d' = abs (n - v)
			

memoise :: [[Int]] -> Memo
memoise = foldl insert empty
insert memo xs = store xs (mkTrees memo xs) memo

countdown :: Int -> [Int] -> (Expr, Value)
countdown n = nearest n . concatMap toExprs . extract . memoise . subseqs

display :: (Expr, Value) -> String
display (e, v)	= (display' e) ++ " = " ++ show v

display' :: Expr -> String
display' (Num x)	= show x
display' (App op e1 e2)	= "(" ++ (display' e1) ++ (disOp op) ++ (display' e2) ++ ")"
  where disOp Add = " + "
	disOp Sub = " - "
	disOp Mul = " * "
	disOp Div = " / "
	
cdp :: Int -> [Int] -> IO()
cdp n ns = print (display (countdown n ns))

exprs :: [Int] -> [(Expr, Value)]	
exprs = concatMap toExprs . extract . memoise . subseqs

--main = print (display (countdown 83100 [1,2,3,4,5,6,7,8]))
--main = print (display (countdown 831 [1,3,7,10,25,50]))